@@ -66,6 +66,12 @@ typedef struct _WnckTaskClass   WnckTaskClass;
 
 #define N_SCREEN_CONNECTIONS 5
 
+#define POINT_IN_RECT(xcoord, ycoord, rect) \
+ ((xcoord) >= (rect).x &&                   \
+  (xcoord) <  ((rect).x + (rect).width) &&  \
+  (ycoord) >= (rect).y &&                   \
+  (ycoord) <  ((rect).y + (rect).height))
+
 typedef enum
 {
   WNCK_TASK_CLASS_GROUP,
@@ -167,6 +173,9 @@ struct _WnckTasklistPrivate
   SnMonitorContext *sn_context;
   guint startup_sequence_timeout;
 #endif
+
+  gint monitor_num;
+  GdkRectangle monitor_geometry;
 };
 
 
@@ -252,14 +261,21 @@ static void     wnck_tasklist_check_end_sequence       (WnckTasklist   *tasklist
 static gpointer task_parent_class;
 static gpointer tasklist_parent_class;
 
+/*
+ * Keep track of all tasklist instances so we can decide
+ * whether to show windows from all monitors in the
+ * tasklist
+ */
+static GSList *tasklist_instances;
+
 /**
- * eel_gtk_label_make_bold.
+ * make_gtk_label_bold.
  *
  * Switches the font of label to a bold equivalent.
  * @label: The label.
  **/
-void
-eel_gtk_label_make_bold (GtkLabel *label)
+static void
+make_gtk_label_bold (GtkLabel *label)
 {
   PangoFontDescription *font_desc;
 
@@ -278,8 +294,8 @@ eel_gtk_label_make_bold (GtkLabel *label)
   pango_font_description_free (font_desc);
 }
 
-void
-wnck_gtk_label_make_normal (GtkLabel *label)
+static void
+make_gtk_label_normal (GtkLabel *label)
 {
   PangoFontDescription *font_desc;
 
@@ -503,9 +519,16 @@ wnck_tasklist_init (WnckTasklist *tasklist)
 
   tasklist->priv->idle_callback_tag = 0;
 
+  tasklist->priv->monitor_num = -1;
+
   atk_obj = gtk_widget_get_accessible (widget);
   atk_object_set_name (atk_obj, _("Window List"));
   atk_object_set_description (atk_obj, _("Tool to switch between visible windows"));
+
+  tasklist_instances = g_slist_append (tasklist_instances, tasklist);
+  g_slist_foreach (tasklist_instances,
+		   (GFunc) wnck_tasklist_update_lists,
+		   NULL);
 }
 
 static void
@@ -534,6 +557,11 @@ wnck_tasklist_finalize (GObject *object)
   WnckTasklist *tasklist;
 
   tasklist = WNCK_TASKLIST (object);
+  
+  tasklist_instances = g_slist_remove (tasklist_instances, tasklist);
+  g_slist_foreach (tasklist_instances,
+		   (GFunc) wnck_tasklist_update_lists,
+		   NULL);
 
   if (tasklist->priv->free_icon_loader_data != NULL)
     (* tasklist->priv->free_icon_loader_data) (tasklist->priv->icon_loader_data);
@@ -1170,6 +1198,8 @@ wnck_tasklist_realize (GtkWidget *widget)
 #endif
   
   (* GTK_WIDGET_CLASS (tasklist_parent_class)->realize) (widget);
+
+  wnck_tasklist_update_lists (tasklist);
 }
 
 static void
@@ -1445,10 +1475,20 @@ static gboolean
 wnck_tasklist_include_window (WnckTasklist *tasklist, WnckWindow *win)
 {
   WnckWorkspace *active_workspace;
+  int x, y, w, h;
 
   if (wnck_window_get_state (win) & WNCK_WINDOW_STATE_SKIP_TASKLIST)
     return FALSE;
 
+  if (tasklist->priv->monitor_num != -1)
+    {
+      wnck_window_get_geometry (win, &x, &y, &w, &h);
+      /* Don't include the window if its center point is not on the same monitor */
+      if (gdk_screen_get_monitor_at_point (_wnck_screen_get_gdk_screen (tasklist->priv->screen),
+                                           x + w / 2, y + h / 2) != tasklist->priv->monitor_num)
+        return FALSE;
+    }
+
   if (tasklist->priv->include_all_workspaces)
     return TRUE;
 
@@ -1480,9 +1520,33 @@ wnck_tasklist_update_lists (WnckTasklist *tasklist)
 
   wnck_tasklist_free_tasks (tasklist);
   
-  windows = wnck_screen_get_windows (tasklist->priv->screen);
-  
-  l = windows;
+  if (GTK_WIDGET (tasklist)->window != NULL)
+    {
+      /*
+       * only show windows from this monitor if there is more than one tasklist running
+       */
+      if (tasklist_instances == NULL || tasklist_instances->next == NULL)
+	{
+	  tasklist->priv->monitor_num = -1;
+        }
+      else
+	{
+	  int monitor_num;
+
+	  monitor_num = gdk_screen_get_monitor_at_window (_wnck_screen_get_gdk_screen (tasklist->priv->screen),
+							  GTK_WIDGET (tasklist)->window);
+
+	  if (monitor_num != tasklist->priv->monitor_num)
+	    {
+	      tasklist->priv->monitor_num = monitor_num;
+	      gdk_screen_get_monitor_geometry (_wnck_screen_get_gdk_screen (tasklist->priv->screen),
+					       tasklist->priv->monitor_num,
+					       &tasklist->priv->monitor_geometry);
+	    }
+	}
+    }
+ 
+  l = windows = wnck_screen_get_windows (tasklist->priv->screen);
   while (l != NULL)
     {
       win = WNCK_WINDOW (l->data);
@@ -1714,18 +1778,39 @@ wnck_tasklist_window_changed_geometry (WnckWindow   *window,
 {
   WnckTask *win_task;
   gboolean show;
+  gboolean monitor_changed;
+  int x, y, w, h;
 
   if (tasklist->priv->idle_callback_tag != 0)
     return;
 
+  /*
+   * If the (parent of the) tasklist itself skips
+   * the tasklist, we need an extra check whether
+   * the tasklist itself possibly changed monitor.
+   */
+  monitor_changed = FALSE;
+  if (wnck_window_get_state (window) & WNCK_WINDOW_STATE_SKIP_TASKLIST &&
+      GTK_WIDGET (tasklist)->window != NULL)
+    {
+      /* Do the extra check only if there is a suspect of a monitor change (= this window is off monitor) */
+      wnck_window_get_geometry (window, &x, &y, &w, &h);
+      if (!POINT_IN_RECT (x + w / 2, y + h / 2, tasklist->priv->monitor_geometry))
+        {
+          monitor_changed = (gdk_screen_get_monitor_at_window (_wnck_screen_get_gdk_screen (tasklist->priv->screen),
+                                                     GTK_WIDGET (tasklist)->window) != tasklist->priv->monitor_num);
+        }
+    }
+
   /*
    * We want to re-generate the task list if
    * the window is shown but shouldn't be or
-   * the window isn't shown but should be.
+   * the window isn't shown but should be or
+   * the tasklist itself changed monitor.
    */
   win_task = g_hash_table_lookup (tasklist->priv->win_hash, window);
   show = wnck_tasklist_include_window(tasklist, window);
-  if ((win_task == NULL && !show) || (win_task != NULL && show))
+  if (((win_task == NULL && !show) || (win_task != NULL && show)) && !monitor_changed)
     return;
 
   /* Don't keep any stale references */
@@ -2178,6 +2263,44 @@ wnck_task_get_icon (WnckTask *task)
   return pixbuf;
 }
 
+static gboolean
+wnck_task_get_demands_attention (WnckTask *task)
+{
+  GList *l;
+  WnckTask *win_task;
+  gboolean demands_attention;
+
+  demands_attention = FALSE;
+
+  switch (task->type)
+    {
+    case WNCK_TASK_CLASS_GROUP:
+      l = task->windows;
+      while (l)
+	{
+	  win_task = WNCK_TASK (l->data);
+
+	  if (wnck_window_get_state (win_task->window) & WNCK_WINDOW_STATE_DEMANDS_ATTENTION)
+	    {
+	      demands_attention = TRUE;
+	      break;
+	    }
+
+	  l = l->next;
+	}
+      break;
+
+    case WNCK_TASK_WINDOW:
+      demands_attention =
+	wnck_window_get_state (task->window) & WNCK_WINDOW_STATE_DEMANDS_ATTENTION;
+      break;
+
+    case WNCK_TASK_STARTUP_SEQUENCE:
+      break;
+    }
+
+  return demands_attention != FALSE;
+}
 
 static void
 wnck_task_update_visible_state (WnckTask *task)
@@ -2195,11 +2318,10 @@ wnck_task_update_visible_state (WnckTask *task)
   if (text != NULL)
     {
       gtk_label_set_text (GTK_LABEL (task->label), text);
-      if (wnck_window_get_state (task->window) &
-          WNCK_WINDOW_STATE_DEMANDS_ATTENTION)
-        eel_gtk_label_make_bold ((GTK_LABEL (task->label)));
+      if (wnck_task_get_demands_attention (task))
+        make_gtk_label_bold ((GTK_LABEL (task->label)));
       else
-        wnck_gtk_label_make_normal ((GTK_LABEL (task->label)));
+        make_gtk_label_normal ((GTK_LABEL (task->label)));
       gtk_tooltips_set_tip (task->tasklist->priv->tooltips, task->button, text, NULL);
       g_free (text);
     }
@@ -2435,9 +2557,12 @@ wnck_task_create_widgets (WnckTask *task)
 
   text = wnck_task_get_text (task);
   task->label = gtk_label_new (text);
-  if (wnck_window_get_state (task->window) &
-      WNCK_WINDOW_STATE_DEMANDS_ATTENTION)
-    eel_gtk_label_make_bold ((GTK_LABEL (task->label)));
+  gtk_misc_set_alignment (GTK_MISC (task->label), 0.0, 0.5);
+  gtk_label_set_ellipsize (GTK_LABEL (task->label),
+                          PANGO_ELLIPSIZE_END);
+
+  if (wnck_task_get_demands_attention (task))
+    make_gtk_label_bold ((GTK_LABEL (task->label)));
   gtk_widget_show (task->label);
 
   gtk_table_attach (GTK_TABLE (table),
@@ -2445,13 +2570,13 @@ wnck_task_create_widgets (WnckTask *task)
 		    0, 1,
 		    0, 1,
 		    0, GTK_EXPAND,
-		    0, 0);
+		    4, 0);
   gtk_table_attach (GTK_TABLE (table),
 		    task->label,
 		    1, 2,
 		    0, 1,
-		    GTK_FILL, GTK_EXPAND,
-		    4, 0);
+		    GTK_FILL | GTK_EXPAND, GTK_EXPAND,		    
+		    2, 0);
 
   gtk_container_add (GTK_CONTAINER (task->button), table);
   gtk_widget_show (table);
@@ -2516,7 +2641,7 @@ draw_dot (GdkWindow *window, GdkGC *lgc, GdkGC *dgc, int x, int y)
 }
 
 
-gboolean
+static gboolean
 wnck_task_class_group_expose (GtkWidget        *widget,
 			      GdkEventExpose   *event,
 			      gpointer          data)
